import os

import curses
import boto3
from colored.colored import bg, fg

from .hud import HUDComponent
from a9s.components.custom_string import String
from a9s.components.table import ColSettings, Table


IS_LOCAL = os.environ.get('LOCAL', 'false').lower() == 'true'


class Route53Table(Table, HUDComponent):
    SERVICE_NAME = 'Route 53'

    def __init__(self) -> None:
        self.client = boto3.client(service_name='route53', endpoint_url='http://localhost:4566' if IS_LOCAL else None)
        self.hosted_zone = None
        self._selection_stack = []
        self._filter_stack = []

        headers, data = self.list_hosted_zones()
        super().__init__(headers, data)

    def get_hud_text(self, space_left):
        if not self.hosted_zone:
            return super().get_hud_text(space_left)

        return String(self.hosted_zone['Name'], bg=bg('medium_purple_2a'), fg=fg('black')).reset_style_on_end()

    def handle_key(self, key):
        should_stop = super().handle_key(key)
        if key.code == curses.KEY_EXIT and not should_stop:
            if self.filter:
                return should_stop

            if self.hosted_zone is not None:
                self.headers, self.data = self.list_hosted_zones()
                self.hosted_zone = None
                should_stop = True

            if len(self._filter_stack) > 0 or len(self._selection_stack) > 0:
                self.filter = self._filter_stack.pop()
                self.selected_row = self._selection_stack.pop()

        return should_stop

    def on_select(self, data):
        if not self.hosted_zone:
            self._filter_stack.append(self.filter)
            self._selection_stack.append(self.selected_row)
            self.hosted_zone = data
            self.headers, self.data = self.list_records()

    def list_hosted_zones(self):
        response = self.client.list_hosted_zones()
        headers = [ColSettings("Name", yank_key='k', stretched=True, min_size=20), ColSettings("ID", yank_key='i'), ColSettings("Records")]

        data = []
        for item in response['HostedZones']:
            data.append([item['Name'], item['Id'], str(item['ResourceRecordSetCount'])])

        return headers, data

    def list_records(self):
        headers = [ColSettings('Name', yank_key='k', min_size=20, stretched=True), ColSettings('Type'), ColSettings('TTL'), ColSettings('Record', yank_key='r')]
        data = []

        should_continue = True
        next_record_name = None
        while should_continue:
            if next_record_name is None:
                response = self.client.list_resource_record_sets(HostedZoneId=self.hosted_zone['ID'], MaxItems='300')
            else:
                response = self.client.list_resource_record_sets(HostedZoneId=self.hosted_zone['ID'], StartRecordName=next_record_name, MaxItems='300')

            if not response['IsTruncated']:
                should_continue = False

            else:
                next_record_name = response['NextRecordName']

            for item in response['ResourceRecordSets']:
                record = item.get('AliasTarget', {}).get('DNSName')
                if not record and item.get('ResourceRecords'):
                    record = ', '.join(r['Value'] for r in item['ResourceRecords'])
                data.append([item['Name'], item['Type'], str(item.get('TTL', '')), record])

        return headers, data



import itertools
from dataclasses import dataclass, field
from typing import Union, Callable, List
from copy import copy

from colored import fg, bg, attr

from .renderer import ScrollableRenderer
from .custom_string import String, Style


@dataclass
class ColSettings:
    name: str
    maxWidth: int = None
    stretched: bool = False
    padding: int = 3
    min_size: Union[int, Callable[[int, "ColSettings"], int]] = None
    max_size: Union[int, Callable[[int, "ColSettings"], int]] = None
    no_shrink: bool = False
    yank_key: str = None

    data_max_len: int = field(init=False, default=0)
    actual_size: int = field(init=False, default=0)
    _min_size: int = 0
    _max_size: int = 0

    def __post_init__(self):
        self.user_defined_min_size = self.min_size is not None
        self.user_defined_max_size = self.max_size is not None

    def enrich_props_using_data(self, col):
        max_data_size = max(len(value) if value else 0 for value in col)
        self.data_max_len = max_data_size

        self._min_size = max(min(len(self.name), max_data_size), 3)
        if self.min_size:
            if callable(self.min_size):
                val = self.min_size(max_data_size, self)

            else:
                val = self.min_size

            self._min_size = max(val, self._min_size)

        self._max_size = max(max_data_size, len(self.name), self._min_size)
        if self.max_size:
            if callable(self.max_size):
                val = self.max_size(max_data_size, self)

            else:
                val = self.max_size

            self._max_size = max(min(val, self._max_size), self._min_size)

        self.actual_size = self._max_size

    def get_full_size(self):
        return self.actual_size + self.padding * 2

    def format_substr(self, data, search, match_style, row_style):
        if len(search) == 0:
            return data

        fragments = data.split(search, 1)
        to_append = String(search, fg=match_style).force_self_style()
        return to_append.join(fragments)

    def format_data(self, data, search="", row_style=attr("reset"), match_style=fg("blue"), yank_mode=False):
        if isinstance(data, str):
            data = String(data)
        
        data = copy(data)

        used_max_size = self.actual_size
        match_index = -1
        if len(search) > 0:
            match_index = data.find(search)

        to_ret = data
        if match_index != -1:
            to_ret = self.format_substr(data, search, match_style, row_style)
        if len(data) > self.actual_size:
            if self.actual_size > self._min_size:
                to_ret = to_ret[:self.actual_size - 3]
                if match_index > self.actual_size - 3 or (match_index != -1 and len(search) > self.actual_size - 3):
                    to_ret += String("...", fg=match_style)
                else:
                    to_ret += "..."

            else:
                if len(data) > self._min_size:
                    to_ret = to_ret[:self._min_size - 3]
                    if match_index > self._min_size - 3 or (match_index != -1 and len(search) > self._min_size - 3):
                        to_ret += String("...", fg=match_style)
                    else:
                        to_ret += "..."
                else:
                    to_ret = to_ret[:self._min_size]

                used_max_size = self._min_size


        if yank_mode and self.yank_key:
            to_ret += String("({})".format(self.yank_key))
    
        data_len = len(to_ret)
        if used_max_size - data_len > 0:
            to_ret += " " * (used_max_size - data_len)

        return to_ret


class Table(ScrollableRenderer):
    def __init__(self, headers: List[ColSettings], data):
        super(Table, self).__init__(data)
        self._transposed_data = []

        self.headers = headers
        self.data = data

    def on_data_set(self, new_value):
        self._transposed_data = list(map(list, itertools.zip_longest(*new_value, fillvalue=None)))
        self.selected_row = 0
        self.offset = 0
        self._displayed_data_start = 0
        for header, col in zip(self.headers, self._transposed_data):
            header.enrich_props_using_data(col)

    def _enrich_reponsive_headers(self):
        stretched_headers = []
        stretched_size = self.width
        stretched_padding = 0
        for header in self.headers:
            if header.stretched:
                stretched_padding += header.padding * 2
                stretched_headers.append(header)
                continue
            stretched_size -= header.get_full_size()

        stretched_size -= stretched_padding
        for header in stretched_headers:
            if header.user_defined_max_size and stretched_size > header._max_size:
                continue

            if header.user_defined_min_size and stretched_size < header._min_size:
                continue

            header.actual_size = stretched_size
        return stretched_size

    @property
    def filtered_data(self):
        if len(self.filter) == 0:
            return self.data

        to_ret = []
        for data_row in self.data:
            found = False
            for col in data_row:
                if self.filter and self.filter in col:
                    found = True
                    break

            if found:
                to_ret.append(data_row)

        return to_ret

    def get_row_representation(self, data):
        to_ret = {}
        for col, d in zip(self.headers, data):
            if isinstance(d, String):
                d = repr(d)
            to_ret[col.name] = d

        return to_ret

    def draw(self, echo):
        header_style = Style(fg=fg("light_gray"), bg=bg("dodger_blue_2"))
        row_to_print = String("").with_style(header_style)
        self._enrich_reponsive_headers()

        for header in self.headers:
            padding = " " * header.padding
            row_to_print += String(padding).with_style(header_style)
            row_to_print += header.format_data(header.name.upper(), yank_mode=self.yank_mode)
            row_to_print += String(padding).with_style(header_style)

        self.max_offset = len(row_to_print) - self.width
        self._curr_row = 0
        row_to_print = self._trim_row(row_to_print, with_elipsis=True)
        echo(self.x, self.y + self._curr_row, row_to_print)
        self._curr_row += 1

        for i, data_row in enumerate(self.displayed_data[self.displayed_data_start:self.displayed_data_end]):
            actual_i = self.displayed_data_start + i
            row_style = Style()
            if self.selected_row == actual_i:
                row_style = Style(fg=fg("black"), bg=bg("light_gray"))

            row_to_print = String("").with_style(row_style)
            for header, data in zip(self.headers, data_row):
                padding = " " * header.padding
                row_to_print += String(padding).with_style(row_style)
                row_to_print += header.format_data(data, self._filter)
                row_to_print += String(padding).with_style(row_style)

            row_to_print = self._trim_row(row_to_print)
            echo(self.x, self.y + self._curr_row, row_to_print)
            self._curr_row += 1
        
        super().draw(echo)



import signal
import curses
import pyperclip

from .aws_resources.logo import Logo
from .aws_resources.hud import HUD
from .aws_resources.services import ServicesSelector
from .components.app import App
from .components.logger import logger
from .components.mode import KeyMode, Mode
from .components.autocomplete import AutoComplete
from . import __version__


class MainApp(App):
    def __init__(self):
        super(MainApp, self).__init__()
        self.logo = Logo(__version__)
        self.hud = HUD()
        self.services_selector = ServicesSelector(hud=self.hud)
        self.mode_renderer = Mode()

        self.auto_complete = AutoComplete(commands=self.commands)
        self.add_multiple([self.logo, self.services_selector, self.mode_renderer, self.auto_complete, logger, self.hud])

        self.on_resize()

    @property
    def commands(self):
        return {
            'debug': self.on_debug_command,
            'quit': self.on_quit_command,

            's3': lambda: self.services_selector.set_service('s3'),
            'route53': lambda: self.services_selector.set_service('route53'),
        }

    def on_debug_command(self):
        self.mode_renderer.mode = KeyMode.Debug
        logger.halt_debug()

    def on_quit_command(self):
        self.should_run = False

    def on_resize(self, *args):
        self.logo.set_pos(x=3, y=0, to_x=24, to_y=6)
        self.hud.set_pos(x=0, y=9, to_x=self.term.width)
        self.mode_renderer.set_pos(x=0, y=self.term.height - 1, to_x=10)
        self.auto_complete.set_pos(x=11, y=self.term.height - 1, to_x=self.term.width)
        self.services_selector.set_pos(x=0, y=10, to_x=self.term.width, to_y=self.term.height - 1)
        logger.set_pos(x=max(self.term.width - 40, 0), y=0, to_x=self.term.width, to_y=10)
        self.services_selector.onresize()
        self.resize()
        self.auto_complete.to_x = self.term.width
        self.clear()

    def run(self):
        self.mode_renderer.mode = KeyMode.Navigation
        for key in self.interactive_run():
            original_mode = self.mode_renderer.mode
            if key:
                logger.debug("pressed key: " + repr(key))

            if key.code == curses.KEY_EXIT:
                self.auto_complete.text = ""

            if original_mode == KeyMode.Navigation:
                should_stop = self.services_selector.handle_key(key)
                self.auto_complete.text = ("/" + self.services_selector.current_service.filter) if self.services_selector.current_service.filter else ""
                if key.code == curses.KEY_EXIT and not should_stop:
                    self.services_selector.current_service.filter = ""

                if key == "/":
                    self.mode_renderer.mode = KeyMode.Search
                    logger.debug("Switching to Search mode")
                elif key == ":":
                    self.mode_renderer.mode = KeyMode.Command
                    logger.debug("Switching to Command mode")

                if key in ["/", ":"]:
                    self.auto_complete.text = str(key)

            if original_mode == KeyMode.Debug:
                if key.code == curses.KEY_EXIT:
                    self.mode_renderer.mode = KeyMode.Navigation
                    logger.continue_debug()

                else:
                    logger.handle_key(key)

            if original_mode in [KeyMode.Search, KeyMode.Command]:
                if key.code == curses.KEY_EXIT:
                    self.mode_renderer.mode = KeyMode.Navigation
                    logger.debug("Switching to Navigate mode")

                if key == "\x16":  # ctrl + v
                    self.auto_complete.text += pyperclip.paste().replace("\n", "").replace("\r", "")

                elif key.code == curses.KEY_BACKSPACE:
                    self.auto_complete.delete_char()

                elif key.code == curses.KEY_ENTER:
                    self.mode_renderer.mode = KeyMode.Navigation
                    logger.debug("Switching to Navigate mode")
                    if self.auto_complete.get_actual_text() == "":
                        self.auto_complete.text = ""

                    self.auto_complete.execute_command_if_matched()

                elif key.isprintable():
                    self.auto_complete.text += key

                if original_mode == KeyMode.Search:
                    self.services_selector.current_service.filter = self.auto_complete.get_actual_text()

                if original_mode == KeyMode.Command:
                    self.auto_complete.handle_key(key)

            if original_mode == KeyMode.Command:
                if key.code == curses.KEY_ENTER:
                    self.auto_complete.text = ""


def main():
    app = MainApp()

    try:
        signal.signal(signal.SIGWINCH, app.on_resize)

    except Exception:
        pass

    app.run()


if __name__ == '__main__':
    main()
